
// Written by Hans de Nivelle, August 2014.
// Revised May 2015.


#include "disjunctive.h"
#include "../tostring.h"


void geometric::disjunctive::insert_rhs( const varatom& at )
{
   if( ispassive( at ))
      throw std::runtime_error( std::string( "cannot insert " ) +
          tostring( at ) + " into rhs, because it is passive" );

   for( const varatom& a : rhs )
   {
      if( a.p == at.p && a.t == at.t && equal_arguments( a, at ))
         return;
   }

   rhs. push_back( at ); 
}


void geometric::disjunctive::set_rhs( )
{
   // Move everything to lhs:

   for( const varatom& a : rhs )
      insert_lhs( a ); 
 
   rhs. clear( );

   auto p = lhs. begin( );  
   while( p != lhs. end( ))
   {
      // If *p is passive, it stays left:

      if( ispassive( *p ))
         ++ p;
      else
      {
         // There is no point in using insert_rhs here, because
         // *p is certainly not passive, and maximally merged.

         switch( p -> t )
         {
         case truthval::E:
         case truthval::T:
            rhs. push_back( *p ); 
            p = lhs. erase( p );
            break; 
         case truthval::ET:
            {
               varatom at = *p;
               at.t = truthval::E;
               rhs. push_back( at );
               at.t = truthval::T;
               rhs. push_back( at );
               p = lhs. erase( p );
            }
            break;
         default:
            throw std::runtime_error( 
               std::string( "set_rhs cannot handle " ) + tostring( *p ));
         }
      }
   }
}


void geometric::disjunctive::makerangerestricted( )
{
   // Make sure that there are #_{F} atoms for every
   // atom in the lhs that needs them:

   formula::makerangerestricted( );

   // Insert #_{F} atoms for all variables in rhs:

   for( const varatom& at : rhs )
   {
      for( variable v : at )
      {
         insert_lhs( varatom( at.p. getcreatefromsignature( ),
                              truthval::F, { v } ));
      }
   }
}


bool geometric::disjunctive::isfree( variable v, 
        std::vector< varatom > :: const_iterator it )  const 
{
   if( formula::isfree( v, it )) 
      return true;

   for( const varatom& at : rhs )
   {
      for( variable w : at )
         if( v == w )
            return true;
   }
   return false;
}


bool geometric::disjunctive::hasothervars( variable v ) const 
{
   if( formula::hasothervars(v))
      return true;

   for( const varatom& at : rhs )
   {
      for( variable w : at )
         if( v != w )
             return true;
   }
   return false;
}


#if 0

geometric::normalizer geometric::disjunctive::normalize( ) 
{
   normalizer norm = formula::normalize( );
   for( auto p = active. begin( ); p != active. end( ); ++ p )
      *p = norm. apply( *p, label(0));
   return norm;
}

#endif


void geometric::disjunctive::print( std::ostream& out ) const 
{
   print_lhs( out ); 

   out << " | "; 
   if( rhs. size( ))
   {
      for( auto p = rhs. begin( ); p != rhs. end( ); ++ p )
      {
         if( p != rhs. begin( ))
            out << ", ";
         out << *p;
      }
   }
   else
      out << "FALSE";
}


#if 0
auto geometric::disjunctive::getfreevariables( ) const -> set<variable>
{
   set<variable> res = formula::getfreevariables( );
   for( const varatom& at : active )
   {
      for( variable v : at )
         res. insert(v);
   }
   return res;
}

#endif 

