
// Written by Hans de Nivelle, August 2014.

#include "existential.h"
#include <stdexcept>


void geometric::existential::check_rhs( ) const 
{
   if( ispassive( at ))
      throw std::runtime_error( std::string( "existential: " ) +
         "quantified atom " + tostring( at ) + " is passive" );

   if( at.t != truthval::E && at.t != truthval::T )
      throw std::runtime_error( std::string( "existential: " ) +
         "quantified atom " + tostring( at ) + " has ambiguous truth value" );

   if( ! at. contains( var ))
      throw std::runtime_error( std::string( "existential: " ) +
        tostring( var ) + " does not occur in " + tostring( at ));
}


void geometric::existential::makerangerestricted( )
{
   formula::makerangerestricted( );

   for( auto p = at. args_cbegin( ); p != at. args_cend( ); ++ p )
   {
      if( *p != var )
      {
         insert_lhs( varatom( at.p. getcreatefromsignature( ), 
                              truthval::F, { *p } ));
      }
   }
}


void geometric::existential::print( std::ostream& out ) const 
{
   print_lhs( out );

   out << " | ";
   out << "EXISTS " << var << ": " << at;
}


bool geometric::existential::hasothervars( variable v ) const
{
   if( formula::hasothervars( v ))
      return true;

   for( auto p = at. args_cbegin( ); p != at. args_cend( ); ++ p )
   {
      if( *p != var && *p != v )
         return true;
   }

   return false; 
}


bool geometric::existential::isfree( variable v, 
         std::vector< varatom > :: const_iterator it ) const
{
   if( formula::isfree( v, it ))
      return true; 

   if( v == var ) 
      return false; 

   for( auto p = at. args_cbegin( ); p != at. args_cend( ); ++ p )
   {
      if( v == *p )
         return true;
   }

   return false;
}


