
// Written by Hans de Nivelle, August 2014.
// Rewritten in May 2015.

#ifndef GEOMETRIC_EXISTENTIAL_INCLUDED
#define GEOMETRIC_EXISTENTIAL_INCLUDED  1

#include "formula.h"

namespace geometric
{

   // An existential rule has form 
   // A_1, ..., A_p | EXISTS y B, where the A_i are inherently passive, and 
   // B is not inherently passive. Variable y must not occur in A_1,...,A_p,
   // but must occur in B. 

   class existential : public formula
   {

      variable var;     // The quantified variable. 
      varatom at;     // The quantified atom. 

   private:
      void check_rhs( ) const;
         // Throw exception if rhs is not well-formed.

   public:  
      existential( variable var, const varatom& at )
         : var{ var }, at{ at } { check_rhs( ); } 
      existential( std::initializer_list< varatom > lhs, 
                   variable var, const varatom& at )
         : formula{ lhs }, var{ var }, at{ at } { check_rhs( ); } 
      existential( const std::vector< varatom > & lhs,
                   variable var, const varatom& at )
         : formula{ lhs }, var{ var }, at{ at } { check_rhs( ); } 

      variable getvariable( ) const { return var; }
      const varatom& getatom( ) const { return at; } 

      void makerangerestricted( ) override;

      void print( std::ostream& out ) const override; 

      existential* clone( ) const & override 
         { return new existential( *this ); } 
      existential* clone( ) && override
         { return new existential( std::move( *this )); }

      bool hasothervars( variable v ) const;
         
      using formula::isfree;  
      bool isfree( variable v, 
                   std::vector< varatom > :: const_iterator it ) const;
      
   };

}

#endif


