
// Written by Hans de Nivelle, January 2015.

#ifndef GEOMETRIC_FALSE_INSTANCE_INCLUDED
#define GEOMETRIC_FALSE_INSTANCE_INCLUDED  1

#include "rule.h" 
#include "subststack.h"
#include "set.h"

namespace geometric 
{ 

   // A false instance contains a rule, the number of choices,
   // an instantiating substitution, and a set of assumptions. 

   struct false_instance
   {
      rule r; 
      unsigned int nrchoices; 
         // Do not assume that nrchoices == rule.nractive( ), because
         // some active atoms may be inconsistent!

      subst< element > inst; 
      set< size_t > assumptions;     // Assumptions that we depend on.

      false_instance( const rule& r,
                      unsigned int nrchoices, 
                      const subst< element > & inst, 
                      const set< size_t > & assumptions )
         : r{r},
           nrchoices{ nrchoices }, 
           inst{ inst },      
           assumptions{ assumptions }
      { 
      }

      // Return the assumption level. It is 0 if we have no assumptions,
      // and n+1 if our highest assumption is n.

      size_t level( ) const
      {
         if( assumptions. size( ) == 0 )
            return 0;
         else
            return assumptions. max( ) + 1;
      }


   };


   bool subsumes( const false_instance& f1, const false_instance& f2 );
      // Subsumption is polynomial, because the instances are present.

   bool isbetterthan( const false_instance& f1, const false_instance& f2 );
      // True if f1 is better than f2 as next candidate for expansion. 

   std::ostream& operator << ( std::ostream& out, const false_instance& f );

}

#endif

 
