
// Written by Hans de Nivelle, Jan 2015.

#include "false_instance_set.h"


void 
geometric::false_instance_set::insert( false_instance&& ff ) 
{
   auto p = lst. begin( );
   while( p != end( ) && ! isbetterthan( ff, *p ))
      ++ p;

   if( p == lst. end( ))
   {
      if( lst. size( ) < maxsize )  
         lst. push_back( std::move(ff));
   }
   else
   {
      lst. insert( p, std::move(ff));  
      if( lst. size( ) > maxsize )
         lst. pop_back( );
   }
}


void
geometric::false_instance_set::insert( false_instance_set&& s )
{
   for( false_instance& ff : s )
      insert( std::move( ff ));
}


void 
geometric::false_instance_set::erase( 
   const std::function< bool( false_instance ) > & f, 
   false_instance_set& erased )
{
   // It seems to me that the best way to do this, is by
   // always moving. At least it is linear, and the moves
   // can be done by splicing. 
   
   false_instance_set kept{ maxsize };
   while( size( ))
   {
      if( f( front( )))
         erased. lst. splice( erased. lst. end( ), lst, begin( ));
      else
         kept. lst. splice( kept. lst. end( ), lst, begin( ));
   }

   *this = std::move( kept );
}


std::ostream& geometric::operator << ( std::ostream& out,
                                       const false_instance_set& best )
{
   out << "False instance set( " << best. maxsize << " ):\n";
   
   size_t nr = 0; 
   for( const auto& f : best )
   {
      out << (nr++) << " : " << f;
   }

   return out;
}


