
#ifndef GEOMETRIC_FALSE_INSTANCE_SET_INCLUDED
#define GEOMETRIC_FALSE_INSTANCE_SET_INCLUDED  1

// A false_instance_set is a set of false_instances.
// Written by Hans de Nivelle, Jan-July 2015. 
// A false_instance_set keeps the best maxsize false
// instances.


#include "false_instance.h"
#include <list>

namespace geometric
{

   struct false_instance_set 
   {
      std::list< false_instance > lst;
      size_t maxsize;

      false_instance_set( size_t maxsize )
         : maxsize{ maxsize }
      { }

      using iterator = std::list< false_instance > :: iterator;
      using const_iterator = std::list< false_instance > :: const_iterator;
 
      iterator begin( ) { return lst. begin( ); }
      iterator end( ) { return lst. end( ); }

      const_iterator cbegin( ) const { return lst. cbegin( ); }
      const_iterator cend( ) const { return lst. cend( ); }

      const_iterator begin( ) const { return lst. cbegin( ); }
      const_iterator end( ) const { return lst. cend( ); }

      void insert( false_instance&& ff );
         // Insert ff at the first position where ff is better than 
         // the current element. 
         // If the set becomes larger than maxsize, the worst 
         // element is deleted. 

      void insert( false_instance_set&& s );
         // Calls the other insert with the members of s.
 
      size_t size( ) const { return lst. size( ); } 

      false_instance& front( ) { return lst. front( ); }
      const false_instance& front( ) const { return lst. front( ); } 

      false_instance& back( ) { return lst. back( ); }
      const false_instance& back( ) const { return lst. back( ); }
 
      void pop_front( ) { lst. pop_front( ); }
      void pop_back( ) { lst. pop_back( ); } 

      iterator erase( iterator it ) { return lst. erase(it); } 

      void erase( const std::function< bool( false_instance ) > & f,
                  false_instance_set& erased );
         // Erase those false_instances for which the function returns true,
         // and append them to erased. 

   };

   std::ostream& operator << ( std::ostream& stream,
                               const false_instance_set& inst );

}

#endif

