
// Written by Hans de Nivelle, Dec. 2014.

#ifndef GEOMETRIC_MATCHING_BASICS_INCLUDED 
#define GEOMETRIC_MATCHING_BASICS_INCLUDED  1

#include "../subststack.h"
#include "../index/simple.h" 
#include "../set.h"


namespace geometric
{

   namespace matching
   {

      // Add the range, corresponding to argument a occurring at position i
      // of an atom that has predicate symbol p.

      template< typename A, typename H, typename E, size_t S >
      inline void 
      addrange( const predicate& p, size_t i, 
                const A& a, const index::simple<A,H,E,S> & ind,
                std::vector< range< size_t >> & res )
      {
         // std::cout << "addrange " << p << ", " << i << ", " << a << " ==> ";

         // We look up the triple (p,i,a) in the index.
         // Arguments of equality must always be looked up with i == 0,
         // because of symmetry.

         range<size_t> r = p.isequality( ) ?  
                                ind.find( p, 0, a ) :
                                ind.find( p, i, a );

               // There are many options here. We can keep all ranges,
               // or keep only the shortest. Currently, we keep only
               // the shortest. One could also keep the two shortest, or
               // whatever.

         // std::cout << "r = " << r << "\n";

         if( res. size( ) == 0 )
            res. push_back(r);
         else
         {
            if( r. size( ) < res. back( ). size( ))
               res. back( ) = r;
         }
      }


      // Returns a vector of ranges that limits the occurrences of
      // atom[ instantiated by subst ], in ind. 

      template< typename A, typename H, typename E, size_t S >
      inline 
      std::vector< range< size_t >> getranges( 
                      const varatom& atom,
                      const subst<A> & subst,
                      const index::simple<A,H,E,S> & ind )
      {
         std::vector< range< size_t >> res;

         for( size_t i = 0; i != atom. getarity( ); ++ i )
         {
            auto val = subst. clookup( atom[i] );
            if( val )
            {
               // If val has content, we look up the triple (p,i,*val):

               addrange( atom.p, i, *val, ind, res );
            }
         }

         // If there was no assigned variable, we simply take all
         // occurrences of the predicate symbol. 

         if( res. size( ) == 0 )
            res. push_back( ind. find( atom.p ));

         return res; 
      }


      // Returns a vector of ranges that limits the occurrences of
      // atom[ instantiated by subst + { v:= a } ] in ind.


      template< typename A, typename H, typename E, size_t S >
      inline 
      std::vector< range< size_t >> getranges( 
                      const varatom& atom,
                      const subst<A> & subst,
                      variable v, const A& a, 
                      const index::simple<A,H,E,S> & ind )
      {
         std::vector< range< size_t >> res;

         for( size_t i = 0; i != atom. getarity( ); ++ i )
         { 
            if( atom[i] == v )
               addrange( atom.p, i, a, ind, res ); 
            else
            { 
               auto val = subst. clookup( atom[i] );
               if( val )
               {
                  addrange( atom.p, i, *val, ind, res );
               }
            }
         }

         // If there was no assigned variable, we take all
         // occurrences of the predicate symbol: 

         if( res. size( ) == 0 )
            res. push_back( ind. find( atom.p ));

         return res; 
      }

   }
}


#endif



