
// Written by Hans de Nivelle, May 2015.

#include "subsumption.h"


std::ostream&
geometric::subsumption::operator << ( std::ostream& out, equality e )
{
   out << e.v1 << " == " << e.v2;
   return out;
}



std::ostream&
geometric::subsumption::operator << ( std::ostream& out,
                                      const skippedequalities& sk )
{
   if( sk.skipped. empty( ))
      out << "complete subsumption\n";
   else
   {
      out << "partial subsumption: {";
      for( auto p = sk. skipped. cbegin( ); p != sk. skipped. cend( ); ++ p )
      {
         if( p == sk. skipped. cbegin( ))
            out << " ";
         else
            out << ", ";
         out << *p;
      }
      out << " }\n";
   }
   return out;
}


void geometric::subsumption::solution(
              const subst< variable > & subst,
              const std::vector< std::pair< variable, variable >> skipped,
              std::vector< skippedequalities > & sol,
              bool& stop )
{

#if 0
   std::cout << "found a solution " << subst << "\n";
#endif

   skippedequalities sk{ skipped, subst };
   // std::cout << sk << "\n\n";

   if( sk. isperfect( ))
      stop = true;

   // The best will be always at the end:

   if( sol. empty( ))
      sol. push_back( sk );
   else
   {
      if( sol. back( ). isworsethan( sk ))
         sol = { sk }; 
   }
}



void geometric::subsumption::subsumes( const disjunctive& disj,
                     size_t level,
                     const index::standardsimple<variable> & index,
                     const set<variable> & variables,
                     subststack<variable> & subst,
                     std::vector<std::pair<variable,variable>> & skipped,
                     std::vector< skippedequalities > & sol,
                     bool& stop )
{

#if 0
   std::cout << "Entering subsumes at level " << level << "\n";
   std::cout << subst << "\n";

   std::cout << "skipped: ";
   for( auto p = skipped. begin( ); p != skipped. end( ); ++ p )
   {
      std::cout << "(" << ( p -> first ) << ", ";
      std::cout << ( p -> second ) << ") ";
   }
   std::cout << "\n";
#endif

   if( level == disj. lhs_size( ) + disj. rhs_size( ))
   {
      solution( subst.s, skipped, sol, stop );
      return;
   }

   const varatom& atom =
      ( level < disj. lhs_size( )) ?
          disj. get_lhs( ) [ level ] :
          disj. get_rhs( ) [ level - disj. lhs_size( ) ];

   if( atom.p.iscreate( ) && atom.t == truthval::F )
   {
      // We very inefficiently enumerate the variables in
      // index.

      if( subst. lookup( atom[0] ))
      {
         throw std::runtime_error( "subsumptionmatch: Argument of "\
                                   "create already has assignment" );
      }
      size_t s = subst. size( );
      for( variable v : variables )
      {
         subst. assign( atom[0], v );
         subsumes( disj, level+1, index, variables,
                   subst, skipped, sol, stop );
         subst. restore(s);
         if( stop ) return;
      }
      return;
   }

   if( atom.p. isequality( ) && atom.t == truthval::T )
   {
      // We know that equality is fully assigned.

      const variable* v1 = subst. lookup( atom[0] );
      const variable* v2 = subst. lookup( atom[1] );
      if( !v1 || !v2 )
      {
         throw std::runtime_error( "subsumptionmatch: Argument of "\
                                   "equality has no assignment" );
      }

      if( *v1 == *v2 )
         return;
            // Because the equality became a tautology.

      std::vector< range< size_t >>
         occ = matching::getranges( atom, subst.s, index );

      bool b = findcommon( occ );
      while(b)
      {
         size_t k = * ( occ[0] ++ );
         const varatom& into = index[k];

         if( into.t == truthval::T )
         {
            if( subst. match( atom[0], into[0] ) &&
                subst. match( atom[1], into[1] ))
            {
               subsumes( disj, level+1, index, variables,
                         subst, skipped, sol, stop );
               return;
            }

            if( subst. match( atom[0], into[1] ) &&
                subst. match( atom[1], into[0] ))
            {
               subsumes( disj, level+1, index, variables,
                         subst, skipped, sol, stop );
               return;
            }
         }
         b = findcommon( occ );
      }

      // If we made it up to here, then we could not match
      // the equality. If we have room left, we skip the equality.

      if( skipped. size( ) < maxskip )
      {
         skipped. push_back( { atom[0], atom[1] } );
         subsumes( disj, level+1, index, variables,
                   subst, skipped, sol, stop );
         skipped. pop_back( );
      }

      return;
   }

   // Usual atom receives usual treatment:

   std::vector< range< size_t >>
      occ = matching::getranges( atom, subst.s, index );

   size_t s = subst. size( );
   bool b = findcommon( occ );
   while(b)
   {
      size_t k = * ( occ[0] ++ );
      const varatom& into = index[k];

      if( ( atom.t & into.t ) == atom.t &&
            subst. match_arguments( atom, into ))
      {
         subsumes( disj, level+1, index, variables,
                   subst, skipped, sol, stop );
      }

      // Restore subst. This must always be done, also when
      // matching failed.

      subst. restore(s);
      if( stop ) return;
      b = findcommon( occ );
   }
}


