
// Written by Hans de Nivelle, May 2015.

#ifndef GEOMETRIC_SUBSUMPTION_INCLUDED 
#define GEOMETRIC_SUBSUMPTION_INCLUDED  1

#include "matching/basics.h" 
#include "subststack.h"
#include "index/simple.h" 
#include "set.h"
#include "disjunctive.h"


namespace geometric
{
   namespace subsumption
   {

      // Used by subsumption:
      // A skipped equality:

      struct equality
      {
         variable v1;
         variable v2;

         equality( variable v1, variable v2 )
            : v1{v1}, v2{v2}
         { }
      };

      std::ostream& operator << ( std::ostream& out, equality e );

      constexpr size_t maxskip = 3;

      // skipped equalities contains equalities that were skipped
      // after the substitution has been applied.
      // If we have F(A,B,C,D), A = B, C = D and
      // F(X,Y,Z,T), then
      // skipped equalities will contain X = Y, Z = T.

      struct skippedequalities
      {
         std::vector< equality > skipped;

         // The constructor applies subst on skipped, and puts
         // the results in our own skipped.

         skippedequalities(
                const std::vector< std::pair<variable,variable>> & skipped,
                const subst< variable > & subst )
         {
            for( auto e : skipped )
            {
               variable v1 = *subst. clookup( e. first );
               variable v2 = *subst. clookup( e. second );
               ( this -> skipped ). push_back( equality( v1, v2 ));
            }
         }

         bool isworsethan( const skippedequalities& other ) const
         {
            return skipped. size( ) > other. skipped. size( );
         }

         bool isperfect( ) const
         {
            return skipped. size( ) == 0;
         }

      };


      std::ostream& operator << ( std::ostream& out,
                                  const skippedequalities & sk );


      void solution(  
              const subst< variable > & subst,
              const std::vector< std::pair< variable, variable >> skipped,
              std::vector< skippedequalities > & sol,
              bool& stop );


      void subsumes( const disjunctive& disj, 
                     size_t level,
                     const index::standardsimple<variable> & index,
                     const set<variable> & variables, 
                     subststack<variable> & subst,
                     std::vector<std::pair<variable,variable>> & skipped, 
                     std::vector< skippedequalities > & sol,
                     bool& stop ); 

   }
}

#endif
