

// Code written by Hans de Nivelle, April 2003. 


#include "inputstream.h"

#include <sstream> 

namespace parser
{

   // If the stream is OK, and we did not reach EOF (which is implied
   //  by the fact that the stream is OK), then we read a character and
   // put it in nextchar.

   void inputstream::moveforward( ) 
   {

   restart:
      if( incl. size( ))
      {
         // Do some line counting:

         if( incl. back( ). nextchar == '\n' )
         {
            ++ incl. back( ). linenumber;
            incl. back( ). columnnumber = 1;
         }
         else
            ++ incl. back( ). columnnumber;

         // Read next character:
 
         incl. back( ). nextchar = incl. back( ). stream. get( );

         // If we reached end of file in the top file, we close it,
         // and are finished.

         if( incl. back( ). stream. eof( ))
         {
            incl. pop_back( );
            return;  
         }

         // If the top file is not good in some other way, 
         // we throw an error message:

         if( !incl. back( ). stream. good( )) 
         {
            std::ostringstream ss;
            ss << "file turned bad ";
            ss << inlineoffile( );
            throw std::runtime_error( ss. str( ));
         }
      }
      else
      {
         if( nextchar == '\n' ) 
         {
            ++ linenumber;
            columnnumber = 1;
         }
         else
            ++ columnnumber;

         if( !stream. good( ))
         {
            std::ostringstream ss;
            ss << "file turned bad ";
            ss << inlineoffile( );
            throw std::runtime_error( ss. str( ));
         }

         nextchar = stream. get( );
      }
   }

   
   void inputstream::include( const std::string& f )
   {
      incl. emplace_back(f);
      if( !incl. back( ). stream )
      {
         std::ostringstream ss;
         ss << "could not open file " << f << " ";
         incl. pop_back( ); 
         ss << inlineoffile( );
         throw std::runtime_error( ss. str( ));
      }

      moveforward( );
   }


   inputstream::~inputstream( ) 
   {
      if( nextchar != EOF ) 
         stream. putback( nextchar );
   }
 
      

   std::string inputstream::inlineoffile( ) const
   {
      std::ostringstream ss;
      ss << "in ";

      if( incl. size( ))
      {
         auto p = incl. end( );
         -- p;
         ss << "line " << ( p -> linenumber );
         ss << " of file " << ( p -> name );

         while( p != incl. begin( ))
         {
            ss << " included from ";
            -- p;
            ss << "line " << ( p -> linenumber );
            ss << " of file " << ( p -> name );
         }
 
         ss << " included from ";
      } 
      ss << "line " << linenumber;
      ss << " of file " << name;
      return ss. str( );
   }

} 



