

// Code written by Hans de Nivelle, April 2003
// In July 2015, I had to make some last minute changes
// for CASC 25. 


#ifndef PARSER_INPUTSTREAM_DECLARED 
#define PARSER_INPUTSTREAM_DECLARED  1

#include <string>
#include <iostream> 
#include <fstream>
#include <stdio.h>
#include <sstream> 
#include <list>
#include <stdexcept>

// An inputstream adds counters (counting line and column) to an 
// istream, it allows for easy inspection of the next character,
// and it is able to stacks of include files. 

namespace parser
{

   struct included
   {
      std::ifstream stream;
      const std::string name;

      unsigned int linenumber;
      unsigned int columnnumber;

      int nextchar;

      included( const std::string& name )
         : stream{ name. c_str( ) },
           name{ name },
           linenumber{ 1 },
           columnnumber{ 1 },
           nextchar{ ' ' }
      { }
   };

 
   class inputstream
   {
      std::istream& stream;
      std::list< included > incl;

      unsigned int linenumber;
      unsigned int columnnumber;
         // Both apply to nextchar. They are intended for use in 
         // error-messages. 

      int nextchar;
         // When the istream is closed, the nextchar is pushed back
         // into the stream.

      const std::string name;
         // Intended for use in error-messages. 

   public: 
      // We maintain the invariant that files that stand at eof,
      // are never kept at incl. 

      bool eof( ) const { return incl. size( ) == 0 && stream. eof( ); } 

   public:
      int getnextchar( ) const
      {
         if( incl. size( ))
            return incl. back( ). nextchar;
         else
            return nextchar;
      }

      void moveforward( );
        // Moveforward makes sure that 
   public:
      inputstream( std::istream& stream = std::cin, 
                   const std::string& name = "STDIN" )
         : stream( stream ),
           linenumber{1},
           columnnumber{0},
           nextchar{ ' ' },
           name{ name } 
      {
         moveforward( ); 
      }
 

      void include( const std::string& f );
         // Include a file name.

      ~inputstream( ); 
         // The next seen character, which was read already from the stream,
         // will be pushed back into the stream, unless it is EOF. 
         // The stream is not closed. 


      inputstream( const inputstream& ) = delete; 
      void operator = ( const inputstream& ) = delete; 

      std::string inlineoffile( ) const;
         // Create the string 'in line L of file F'.
         // This is useful for error messages. 
   };

} 


#endif


