
#ifndef UNIQUE_NUMBER_INCLUDED
#define UNIQUE_NUMBER_INCLUDED  1

// If you want every T to have a unique number,
// give it a field unique_number<T>.
//
// Written by Hans de Nivelle, April 2015. 

#include <iostream>
#include <mutex>


template< typename T > 
struct unique_number
{
   unsigned int value;

   static bool hex;
      // Make it true if you want hexadecimal output.

   // Default constructor puts a unique number in
   // value.

   unique_number( )
   {
      // We cannot use initializers, because that would be
      // not thread safe. We want unique numbers in different
      // threads.

      static unsigned int nr = 0;
      static std::mutex m; 
      m. lock( );
      value = ( nr ++ );
      m. unlock( );
   }
};

template< typename T >
bool unique_number<T>::hex = false;


template< typename T > 
std::ostream& operator << ( std::ostream& out, unique_number<T> u )
{
   if( unique_number<T> :: hex ) 
   {
      out << std::hex; 
      out << std::uppercase;
   }

   out << u. value;

   if( unique_number<T> :: hex ) 
   {
      out << std::nouppercase;
      out << std::dec;
   }

   return out; 
}

#endif
 
